import ast
import collections
import dataclasses
import logging
import pathlib
import typing

from odd.artifact import Artifact
from odd.const import MANIFEST_FILENAMES
from odd.typedefs import OdooVersion

_LOG = logging.getLogger(__name__)


@dataclasses.dataclass(unsafe_hash=True)
class ManifestPath:
    path: pathlib.Path

    @property
    def addon_path(self) -> pathlib.Path:
        return self.path.parent

    @property
    def name(self) -> str:
        return self.addon_path.name


@dataclasses.dataclass
class Addon(Artifact):
    manifest_path: ManifestPath
    manifest: typing.Dict[str, typing.Any]
    version: OdooVersion

    @property
    def path(self) -> pathlib.Path:
        return self.manifest_path.addon_path

    @property
    def name(self) -> str:
        return self.manifest_path.name

    @property
    def data_files(self):
        return [self.path / fn for fn in (self.manifest.get("data") or [])]

    @property
    def demo_files(self):
        return [self.path / fn for fn in (self.manifest.get("demo") or [])]

    @property
    def qweb_files(self):
        files = []
        for fn in self.manifest.get("qweb") or []:
            fn_path = self.path / fn
            if "*" in fn_path.stem:
                files.extend(fn_path.parent.glob(fn_path.name))
            else:
                files.append(fn_path)
        return files


def parse_manifest(manifest_path: ManifestPath) -> typing.Any:
    # FIXME: Check for manifest file size.
    with manifest_path.path.open(mode="r") as f:
        return ast.literal_eval(f.read())


def find_manifest(path: pathlib.Path) -> typing.Optional[pathlib.Path]:
    for child in path.iterdir():
        if child.is_file() and child.name in MANIFEST_FILENAMES:
            return child
    return None


def discover_addons(
    dir_path: pathlib.Path, allow_duplicates: bool = False
) -> typing.Generator[ManifestPath, None, None]:
    """
    :param allow_duplicates: If False, addons with the same name are ignored.
    """
    dirs, seen = collections.deque((dir_path,)), set()

    def check_dir(
        path: pathlib.Path, descend: bool = True
    ) -> typing.Generator[ManifestPath, None, None]:
        manifest = find_manifest(path)
        if manifest:
            addon_name = manifest.parent.name
            if allow_duplicates or addon_name not in seen:
                yield ManifestPath(manifest)
            if not allow_duplicates:
                if addon_name in seen:
                    _LOG.warning(
                        'Addon with the same name "%s" found at path: %s, '
                        "it will be ignored",
                        addon_name,
                        path,
                    )
                else:
                    seen.add(addon_name)
        else:
            if descend:
                dirs.append(path)

    yield from check_dir(dir_path, descend=False)
    while dirs:
        dir = dirs.popleft()
        for child in dir.iterdir():
            if child.is_dir():
                yield from check_dir(child)
