import collections
import pathlib
import typing

from odd.utils import Position

import yaml


class YamlTag(collections.abc.Mapping):
    def __init__(self, tag: str, start_pos: Position, end_pos: Position, **kwargs):
        self.tag = tag
        self.start_pos, self.end_pos = start_pos, end_pos
        self._kwargs = kwargs

    def __getitem__(self, key):
        return self._kwargs[key]

    def __iter__(self):
        return iter(self._kwargs)

    def __len__(self):
        return len(self._kwargs)

    def __hash__(self):
        return hash(repr(self))

    def __repr__(self):
        return f"<!{self.tag} {sorted(self.items())}>"


def line_column_index_1(position: Position) -> Position:
    return Position(position[0] + 1, position[1] + 1)


def _get_positions(node: yaml.Node) -> typing.Tuple[Position, Position]:
    return (
        Position(node.start_mark.line, node.start_mark.column),
        Position(node.end_mark.line, node.end_mark.column),
    )


def _make_generic_constructor(
    tag: str
) -> typing.Callable[[yaml.BaseLoader, yaml.Node], YamlTag]:
    def constructor(loader: yaml.BaseLoader, node: yaml.Node) -> YamlTag:
        return YamlTag(tag, *_get_positions(node), **loader.construct_mapping(node))

    return constructor


assert_constructor = _make_generic_constructor("assert")
record_constructor = _make_generic_constructor("record")
python_constructor = _make_generic_constructor("python")
menuitem_constructor = _make_generic_constructor("menuitem")
workflow_constructor = _make_generic_constructor("workflow")
act_window_constructor = _make_generic_constructor("act_window")
function_constructor = _make_generic_constructor("function")
report_constructor = _make_generic_constructor("report")
delete_constructor = _make_generic_constructor("delete")
context_constructor = _make_generic_constructor("context")
url_constructor = _make_generic_constructor("url")
ir_set_constructor = _make_generic_constructor("ir_set")


def eval_constructor(loader: yaml.BaseLoader, node: yaml.Node) -> YamlTag:
    expression = loader.construct_scalar(node)
    return YamlTag("eval", *_get_positions(node), expression=expression)


def ref_constructor(
    loader: yaml.BaseLoader, tag_suffix: str, node: yaml.Node
) -> YamlTag:
    if tag_suffix == "id":
        kwargs = {"id": loader.construct_scalar(node)}
    else:
        kwargs = loader.construct_mapping(node)
    return YamlTag("ref", *_get_positions(node), **kwargs)


def add_constructors() -> None:
    yaml.add_constructor("!assert", assert_constructor)
    yaml.add_constructor("!record", record_constructor)
    yaml.add_constructor("!python", python_constructor)
    yaml.add_constructor("!menuitem", menuitem_constructor)
    yaml.add_constructor("!workflow", workflow_constructor)
    yaml.add_constructor("!act_window", act_window_constructor)
    yaml.add_constructor("!function", function_constructor)
    yaml.add_constructor("!report", report_constructor)
    yaml.add_constructor("!context", context_constructor)
    yaml.add_constructor("!delete", delete_constructor)
    yaml.add_constructor("!url", url_constructor)
    yaml.add_constructor("!eval", eval_constructor)
    yaml.add_multi_constructor("!ref", ref_constructor)
    yaml.add_constructor("!ir_set", ir_set_constructor)


def load(path: pathlib.Path) -> typing.Any:
    add_constructors()
    with path.open(mode="rb") as f:
        return yaml.load(f, Loader=yaml.Loader)
